@@ -11,8 +11,6 @@ import sys
 import time
 import threading
 
-import psutil
-
 from snakemake.exceptions import WorkflowError
 
 
@@ -39,6 +37,7 @@ class BenchmarkRecord:
                 "io_in",
                 "io_out",
                 "mean_load",
+                "cpu_time",
             )
         )
 
@@ -51,7 +50,8 @@ class BenchmarkRecord:
         max_pss=None,
         io_in=None,
         io_out=None,
-        cpu_seconds=None,
+        cpu_usages=None,
+        cpu_time=None,
     ):
         #: Running time in seconds
         self.running_time = running_time
@@ -68,7 +68,9 @@ class BenchmarkRecord:
         #: I/O written in bytes
         self.io_out = io_out
         #: Count of CPU seconds, divide by running time to get mean load estimate
-        self.cpu_seconds = cpu_seconds or 0
+        self.cpu_usages = cpu_usages or 0
+        #: CPU usage (user and system) in seconds
+        self.cpu_time = cpu_time or 0
         #: First time when we measured CPU load, for estimating total running time
         self.first_time = None
         #: Previous point when measured CPU load, for estimating total running time
@@ -111,7 +113,8 @@ class BenchmarkRecord:
                     self.max_pss,
                     self.io_in,
                     self.io_out,
-                    100.0 * self.cpu_seconds / self.running_time,
+                    self.cpu_usages / self.running_time,
+                    self.cpu_time,
                 ),
             )
         )
@@ -187,6 +190,8 @@ class BenchmarkTimer(ScheduledPeriodicTimer):
     """Allows easy observation of a given PID for resource usage"""
 
     def __init__(self, pid, bench_record, interval=BENCHMARK_INTERVAL):
+        import psutil
+
         ScheduledPeriodicTimer.__init__(self, interval)
         #: PID of observed process
         self.pid = pid
@@ -198,6 +203,8 @@ class BenchmarkTimer(ScheduledPeriodicTimer):
 
     def work(self):
         """Write statistics"""
+        import psutil
+
         try:
             self._update_record()
         except psutil.NoSuchProcess:
@@ -207,13 +214,18 @@ class BenchmarkTimer(ScheduledPeriodicTimer):
 
     def _update_record(self):
         """Perform the actual measurement"""
+        import psutil
+
         # Memory measurements
         rss, vms, uss, pss = 0, 0, 0, 0
         # I/O measurements
         io_in, io_out = 0, 0
         check_io = True
         # CPU seconds
-        cpu_seconds = 0
+        cpu_usages = 0
+        # CPU usage time
+        cpu_time = 0
+
         # Iterate over process and all children
         try:
             this_time = time.time()
@@ -221,16 +233,16 @@ class BenchmarkTimer(ScheduledPeriodicTimer):
                 proc = self.procs.setdefault(proc.pid, proc)
                 with proc.oneshot():
                     if self.bench_record.prev_time:
-                        cpu_seconds += (
-                            proc.cpu_percent()
-                            / 100
-                            * (this_time - self.bench_record.prev_time)
+                        cpu_usages += proc.cpu_percent() * (
+                            this_time - self.bench_record.prev_time
                         )
+
                     meminfo = proc.memory_full_info()
                     rss += meminfo.rss
                     vms += meminfo.vms
                     uss += meminfo.uss
                     pss += meminfo.pss
+
                     if check_io:
                         try:
                             ioinfo = proc.io_counters()
@@ -239,29 +251,40 @@ class BenchmarkTimer(ScheduledPeriodicTimer):
                         except NotImplementedError as nie:
                             # OS doesn't track IO
                             check_io = False
+
+                    cpu_times = proc.cpu_times()
+                    cpu_time += cpu_times.user + cpu_times.system
+
             self.bench_record.prev_time = this_time
             if not self.bench_record.first_time:
                 self.bench_record.prev_time = this_time
+
             rss /= 1024 * 1024
             vms /= 1024 * 1024
             uss /= 1024 * 1024
             pss /= 1024 * 1024
+
             if check_io:
                 io_in /= 1024 * 1024
                 io_out /= 1024 * 1024
             else:
                 io_in = None
                 io_out = None
+
         except psutil.Error as e:
             return
+
         # Update benchmark record's RSS and VMS
         self.bench_record.max_rss = max(self.bench_record.max_rss or 0, rss)
         self.bench_record.max_vms = max(self.bench_record.max_vms or 0, vms)
         self.bench_record.max_uss = max(self.bench_record.max_uss or 0, uss)
         self.bench_record.max_pss = max(self.bench_record.max_pss or 0, pss)
+
         self.bench_record.io_in = io_in
         self.bench_record.io_out = io_out
-        self.bench_record.cpu_seconds += cpu_seconds
+
+        self.bench_record.cpu_usages += cpu_usages
+        self.bench_record.cpu_time = cpu_time
 
 
 @contextlib.contextmanager
