@@ -1,6 +1,6 @@
 __author__ = "Johannes Köster"
-__copyright__ = "Copyright 2015-2019, Johannes Köster"
-__email__ = "koester@jimmy.harvard.edu"
+__copyright__ = "Copyright 2021, Johannes Köster"
+__email__ = "johannes.koester@uni-due.de"
 __license__ = "MIT"
 
 import os
@@ -66,6 +66,7 @@ class AbstractExecutor:
         printthreads=True,
         latency_wait=3,
         keepincomplete=False,
+        keepmetadata=True,
     ):
         self.workflow = workflow
         self.dag = dag
@@ -75,6 +76,7 @@ class AbstractExecutor:
         self.printthreads = printthreads
         self.latency_wait = latency_wait
         self.keepincomplete = keepincomplete
+        self.keepmetadata = keepmetadata
 
     def get_default_remote_provider_args(self):
         if self.workflow.default_remote_provider:
@@ -86,8 +88,39 @@ class AbstractExecutor:
             )
         return ""
 
-    def get_default_resources_args(self):
-        if self.workflow.default_resources.args is not None:
+    def _format_key_value_args(self, flag, kwargs):
+        if kwargs:
+            return " {} {} ".format(
+                flag,
+                " ".join("{}={}".format(key, value) for key, value in kwargs.items()),
+            )
+        return ""
+
+    def get_set_threads_args(self):
+        return self._format_key_value_args(
+            "--set-threads", self.workflow.overwrite_threads
+        )
+
+    def get_set_resources_args(self):
+        if self.workflow.overwrite_resources:
+            return " --set-resources {} ".format(
+                " ".join(
+                    "{}:{}={}".format(rule, name, value)
+                    for rule, res in self.workflow.overwrite_resources.items()
+                    for name, value in res.items()
+                ),
+            )
+        return ""
+
+    def get_set_scatter_args(self):
+        return self._format_key_value_args(
+            "--set-scatter", self.workflow.overwrite_scatter
+        )
+
+    def get_default_resources_args(self, default_resources=None):
+        if default_resources is None:
+            default_resources = self.workflow.default_resources
+        if default_resources:
 
             def fmt(res):
                 if isinstance(res, str):
@@ -100,6 +133,11 @@ class AbstractExecutor:
             return args
         return ""
 
+    def get_behavior_args(self):
+        if self.workflow.conda_not_block_search_path_envvars:
+            return " --conda-not-block-search-path-envvars "
+        return ""
+
     def run_jobs(self, jobs, callback=None, submit_callback=None, error_callback=None):
         """Run a list of jobs that is ready at a given point in time.
 
@@ -182,6 +220,7 @@ class RealExecutor(AbstractExecutor):
         latency_wait=3,
         assume_shared_fs=True,
         keepincomplete=False,
+        keepmetadata=False,
     ):
         super().__init__(
             workflow,
@@ -191,10 +230,11 @@ class RealExecutor(AbstractExecutor):
             printshellcmds=printshellcmds,
             latency_wait=latency_wait,
             keepincomplete=keepincomplete,
+            keepmetadata=keepmetadata,
         )
         self.assume_shared_fs = assume_shared_fs
         self.stats = Stats()
-        self.snakefile = workflow.snakefile
+        self.snakefile = workflow.main_snakefile
 
     def register_job(self, job):
         job.register()
@@ -229,6 +269,7 @@ class RealExecutor(AbstractExecutor):
             ignore_missing_output=ignore_missing_output,
             latency_wait=self.latency_wait,
             assume_shared_fs=self.assume_shared_fs,
+            keep_metadata=self.keepmetadata,
         )
         self.stats.report_job_end(job)
 
@@ -254,6 +295,10 @@ class RealExecutor(AbstractExecutor):
             additional += " --use-conda "
             if self.workflow.conda_prefix:
                 additional += " --conda-prefix {} ".format(self.workflow.conda_prefix)
+            if self.workflow.conda_base_path and self.assume_shared_fs:
+                additional += " --conda-base-path {} ".format(
+                    self.workflow.conda_base_path
+                )
         if self.workflow.use_singularity:
             additional += " --use-singularity "
             if self.workflow.singularity_prefix:
@@ -264,9 +309,21 @@ class RealExecutor(AbstractExecutor):
                 additional += ' --singularity-args "{}"'.format(
                     self.workflow.singularity_args
                 )
+        if not self.workflow.execute_subworkflows:
+            additional += " --no-subworkflows "
+
+        if self.workflow.max_threads is not None:
+            additional += " --max-threads {} ".format(self.workflow.max_threads)
+
+        additional += self.get_set_resources_args()
+        additional += self.get_set_scatter_args()
+        additional += self.get_set_threads_args()
+        additional += self.get_behavior_args()
 
         if self.workflow.use_env_modules:
-            additional += " --use-envmodules"
+            additional += " --use-envmodules "
+        if not self.keepmetadata:
+            additional += " --drop-metadata "
 
         return additional
 
@@ -307,7 +364,7 @@ class RealExecutor(AbstractExecutor):
         if "cores" in kwargs:
             del kwargs["cores"]
 
-        return format(
+        cmd = format(
             pattern,
             job=job,
             attempt=job.attempt,
@@ -320,8 +377,9 @@ class RealExecutor(AbstractExecutor):
             benchmark_repeats=job.benchmark_repeats if not job.is_group() else None,
             target=target,
             rules=rules,
-            **kwargs
+            **kwargs,
         )
+        return cmd
 
 
 class TouchExecutor(RealExecutor):
@@ -361,6 +419,7 @@ class CPUExecutor(RealExecutor):
         latency_wait=3,
         cores=1,
         keepincomplete=False,
+        keepmetadata=True,
     ):
         super().__init__(
             workflow,
@@ -370,15 +429,17 @@ class CPUExecutor(RealExecutor):
             printshellcmds=printshellcmds,
             latency_wait=latency_wait,
             keepincomplete=keepincomplete,
+            keepmetadata=keepmetadata,
         )
 
         self.exec_job = "\\\n".join(
             (
                 "cd {workflow.workdir_init} && ",
                 "{sys.executable} -m snakemake {target} --snakefile {snakefile} ",
-                "--force -j{cores} --keep-target-files --keep-remote ",
-                "--attempt {attempt} ",
+                "--force --cores {cores} --keep-target-files --keep-remote ",
+                "--attempt {attempt} --scheduler {workflow.scheduler_type} ",
                 "--force-use-threads --wrapper-prefix {workflow.wrapper_prefix} ",
+                "--max-inventory-time 0 --ignore-incomplete ",
                 "--latency-wait {latency_wait} ",
                 self.get_default_remote_provider_args(),
                 self.get_default_resources_args(),
@@ -451,7 +512,10 @@ class CPUExecutor(RealExecutor):
             self.workflow.cleanup_scripts,
             job.shadow_dir,
             job.jobid,
-            self.workflow.edit_notebook,
+            self.workflow.edit_notebook if self.dag.is_edit_notebook_job(job) else None,
+            self.workflow.conda_base_path,
+            job.rule.basedir,
+            self.workflow.sourcecache.runtime_cache_path,
         )
 
     def run_single_job(self, job):
@@ -578,6 +642,7 @@ class ClusterExecutor(RealExecutor):
         disable_default_remote_provider_args=False,
         disable_get_default_resources_args=False,
         keepincomplete=False,
+        keepmetadata=True,
     ):
         from ratelimiter import RateLimiter
 
@@ -590,11 +655,13 @@ class ClusterExecutor(RealExecutor):
             printshellcmds=printshellcmds,
             latency_wait=latency_wait,
             assume_shared_fs=assume_shared_fs,
+            keepincomplete=keepincomplete,
+            keepmetadata=keepmetadata,
         )
 
         if not self.assume_shared_fs:
             # use relative path to Snakefile
-            self.snakefile = os.path.relpath(workflow.snakefile)
+            self.snakefile = os.path.relpath(workflow.main_snakefile)
 
         jobscript = workflow.jobscript
         if jobscript is None:
@@ -618,12 +685,12 @@ class ClusterExecutor(RealExecutor):
                     else "",
                     "{sys.executable} " if assume_shared_fs else "python ",
                     "-m snakemake {target} --snakefile {snakefile} ",
-                    "--force -j{cores} --keep-target-files --keep-remote ",
-                    "--wait-for-files {wait_for_files} --latency-wait {latency_wait} ",
-                    " --attempt {attempt} {use_threads} ",
+                    "--force --cores {cores} --keep-target-files --keep-remote --max-inventory-time 0 ",
+                    "{waitfiles_parameter:u} --latency-wait {latency_wait} ",
+                    " --attempt {attempt} {use_threads} --scheduler {workflow.scheduler_type} ",
                     "--wrapper-prefix {workflow.wrapper_prefix} ",
                     "{overwrite_workdir} {overwrite_config} {printshellcmds} {rules} "
-                    "--nocolor --notemp --no-hooks --nolock ",
+                    "--nocolor --notemp --no-hooks --nolock {scheduler_solver_path:u} ",
                     "--mode {} ".format(Mode.cluster),
                 )
             )
@@ -638,7 +705,7 @@ class ClusterExecutor(RealExecutor):
 
         self.jobname = jobname
         self._tmpdir = None
-        self.cores = cores if cores else ""
+        self.cores = cores if cores else "all"
         self.cluster_config = cluster_config if cluster_config else dict()
 
         self.restart_times = restart_times
@@ -646,7 +713,7 @@ class ClusterExecutor(RealExecutor):
         self.active_jobs = list()
         self.lock = threading.Lock()
         self.wait = True
-        self.wait_thread = threading.Thread(target=self._wait_for_jobs)
+        self.wait_thread = threading.Thread(target=self._wait_thread)
         self.wait_thread.daemon = True
         self.wait_thread.start()
 
@@ -656,6 +723,12 @@ class ClusterExecutor(RealExecutor):
             max_calls=self.max_status_checks_per_second, period=1
         )
 
+    def _wait_thread(self):
+        try:
+            self._wait_for_jobs()
+        except Exception as e:
+            self.workflow.scheduler.executor_error_callback(e)
+
     def shutdown(self):
         with self.lock:
             self.wait = False
@@ -695,17 +768,43 @@ class ClusterExecutor(RealExecutor):
 
     def format_job(self, pattern, job, **kwargs):
         wait_for_files = []
+        scheduler_solver_path = ""
         if self.assume_shared_fs:
             wait_for_files.append(self.tmpdir)
             wait_for_files.extend(job.get_wait_for_files())
+            # Prepend PATH of current python executable to PATH.
+            # This way, we ensure that the snakemake process in the cluster node runs
+            # in the same environment as the current process.
+            # This is necessary in order to find the pulp solver backends (e.g. coincbc).
+            scheduler_solver_path = "--scheduler-solver-path {}".format(
+                os.path.dirname(sys.executable)
+            )
+
+        # Only create extra file if we have more than 20 input files.
+        # This should not require the file creation in most cases.
+        if len(wait_for_files) > 20:
+            wait_for_files_file = self.get_jobscript(job) + ".waitforfilesfile.txt"
+            with open(wait_for_files_file, "w") as fd:
+                fd.write("\n".join(wait_for_files))
+
+            waitfiles_parameter = format(
+                "--wait-for-files-file {wait_for_files_file}",
+                wait_for_files_file=repr(wait_for_files_file),
+            )
+        else:
+            waitfiles_parameter = format(
+                "--wait-for-files {wait_for_files}",
+                wait_for_files=[repr(f) for f in wait_for_files],
+            )
 
         format_p = partial(
             self.format_job_pattern,
             job=job,
             properties=job.properties(cluster=self.cluster_params(job)),
             latency_wait=self.latency_wait,
-            wait_for_files=wait_for_files,
-            **kwargs
+            waitfiles_parameter=waitfiles_parameter,
+            scheduler_solver_path=scheduler_solver_path,
+            **kwargs,
         )
         try:
             return format_p(pattern)
@@ -730,13 +829,13 @@ class ClusterExecutor(RealExecutor):
             _quote_all=True,
             use_threads=use_threads,
             envvars=envvars,
-            **kwargs
+            **kwargs,
         )
         content = self.format_job(self.jobscript, job, exec_job=exec_job, **kwargs)
         logger.debug("Jobscript:\n{}".format(content))
         with open(jobscript, "w") as f:
             print(content, file=f)
-        os.chmod(jobscript, os.stat(jobscript).st_mode | stat.S_IXUSR)
+        os.chmod(jobscript, os.stat(jobscript).st_mode | stat.S_IXUSR | stat.S_IRUSR)
 
     def cluster_params(self, job):
         """Return wildcards object for job from cluster_config."""
@@ -783,6 +882,10 @@ class ClusterExecutor(RealExecutor):
         # By removing it again, we make sure that it is gone on the host FS.
         if not self.keepincomplete:
             self.workflow.persistence.cleanup(job)
+            # Also cleanup the jobs output files, in case the remote job
+            # was not able to, due to e.g. timeout.
+            logger.debug("Cleanup failed jobs output files.")
+            job.cleanup()
 
     def print_cluster_job_error(self, job_info, jobid):
         job = job_info.job
@@ -824,6 +927,7 @@ class GenericClusterExecutor(ClusterExecutor):
         assume_shared_fs=True,
         max_status_checks_per_second=1,
         keepincomplete=False,
+        keepmetadata=True,
     ):
 
         self.submitcmd = submitcmd
@@ -849,6 +953,8 @@ class GenericClusterExecutor(ClusterExecutor):
             restart_times=restart_times,
             assume_shared_fs=assume_shared_fs,
             max_status_checks_per_second=max_status_checks_per_second,
+            keepincomplete=keepincomplete,
+            keepmetadata=keepmetadata,
         )
 
         if statuscmd:
@@ -967,21 +1073,18 @@ class GenericClusterExecutor(ClusterExecutor):
         success = "success"
         failed = "failed"
         running = "running"
+        status_cmd_kills = []
         if self.statuscmd is not None:
 
-            def job_status(job):
+            def job_status(job, valid_returns=["running", "success", "failed"]):
                 try:
                     # this command shall return "success", "failed" or "running"
-                    return (
-                        subprocess.check_output(
-                            "{statuscmd} {jobid}".format(
-                                jobid=job.jobid, statuscmd=self.statuscmd
-                            ),
-                            shell=True,
-                        )
-                        .decode()
-                        .split("\n")[0]
-                    )
+                    ret = subprocess.check_output(
+                        "{statuscmd} {jobid}".format(
+                            jobid=job.jobid, statuscmd=self.statuscmd
+                        ),
+                        shell=True,
+                    ).decode()
                 except subprocess.CalledProcessError as e:
                     if e.returncode < 0:
                         # Ignore SIGINT and all other issues due to signals
@@ -989,14 +1092,32 @@ class GenericClusterExecutor(ClusterExecutor):
                         # Ctrl-C on the main process or sending killall to
                         # snakemake.
                         # Snakemake will handle the signal in
-                        # the master process.
-                        pass
+                        # the main process.
+                        status_cmd_kills.append(-e.returncode)
+                        if len(status_cmd_kills) > 10:
+                            logger.info(
+                                "Cluster status command {} was killed >10 times with signal(s) {} "
+                                "(if this happens unexpectedly during your workflow execution, "
+                                "have a closer look.).".format(
+                                    self.statuscmd, ",".join(status_cmd_kills)
+                                )
+                            )
+                            status_cmd_kills.clear()
                     else:
                         raise WorkflowError(
                             "Failed to obtain job status. "
                             "See above for error message."
                         )
 
+                ret = ret.strip().split("\n")
+                if len(ret) != 1 or ret[0] not in valid_returns:
+                    raise WorkflowError(
+                        "Cluster status command {} returned {} but just a single line with one of {} is expected.".format(
+                            self.statuscmd, "\\n".join(ret), ",".join(valid_returns)
+                        )
+                    )
+                return ret[0]
+
         else:
 
             def job_status(job):
@@ -1017,7 +1138,7 @@ class GenericClusterExecutor(ClusterExecutor):
                 active_jobs = self.active_jobs
                 self.active_jobs = list()
                 still_running = list()
-            logger.debug("Checking status of {} jobs.".format(len(active_jobs)))
+            # logger.debug("Checking status of {} jobs.".format(len(active_jobs)))
             for active_job in active_jobs:
                 with self.status_rate_limiter:
                     status = job_status(active_job)
@@ -1069,6 +1190,7 @@ class SynchronousClusterExecutor(ClusterExecutor):
         restart_times=0,
         assume_shared_fs=True,
         keepincomplete=False,
+        keepmetadata=True,
     ):
         super().__init__(
             workflow,
@@ -1083,6 +1205,8 @@ class SynchronousClusterExecutor(ClusterExecutor):
             restart_times=restart_times,
             assume_shared_fs=assume_shared_fs,
             max_status_checks_per_second=10,
+            keepincomplete=keepincomplete,
+            keepmetadata=keepmetadata,
         )
         self.submitcmd = submitcmd
         self.external_jobid = dict()
@@ -1178,6 +1302,7 @@ class DRMAAExecutor(ClusterExecutor):
         assume_shared_fs=True,
         max_status_checks_per_second=1,
         keepincomplete=False,
+        keepmetadata=True,
     ):
         super().__init__(
             workflow,
@@ -1192,6 +1317,8 @@ class DRMAAExecutor(ClusterExecutor):
             restart_times=restart_times,
             assume_shared_fs=assume_shared_fs,
             max_status_checks_per_second=max_status_checks_per_second,
+            keepincomplete=keepincomplete,
+            keepmetadata=keepmetadata,
         )
         try:
             import drmaa
@@ -1277,6 +1404,8 @@ class DRMAAExecutor(ClusterExecutor):
     def _wait_for_jobs(self):
         import drmaa
 
+        suspended_msg = set()
+
         while True:
             with self.lock:
                 if not self.wait:
@@ -1287,9 +1416,7 @@ class DRMAAExecutor(ClusterExecutor):
             for active_job in active_jobs:
                 with self.status_rate_limiter:
                     try:
-                        retval = self.session.wait(
-                            active_job.jobid, drmaa.Session.TIMEOUT_NO_WAIT
-                        )
+                        retval = self.session.jobStatus(active_job.jobid)
                     except drmaa.ExitTimeoutException as e:
                         # job still active
                         still_running.append(active_job)
@@ -1302,20 +1429,40 @@ class DRMAAExecutor(ClusterExecutor):
                         os.remove(active_job.jobscript)
                         active_job.error_callback(active_job.job)
                         continue
-                    # job exited
-                    os.remove(active_job.jobscript)
-                    if (
-                        not retval.wasAborted
-                        and retval.hasExited
-                        and retval.exitStatus == 0
-                    ):
+                    if retval == drmaa.JobState.DONE:
+                        os.remove(active_job.jobscript)
                         active_job.callback(active_job.job)
-                    else:
+                    elif retval == drmaa.JobState.FAILED:
+                        os.remove(active_job.jobscript)
                         self.print_job_error(active_job.job)
                         self.print_cluster_job_error(
                             active_job, self.dag.jobid(active_job.job)
                         )
                         active_job.error_callback(active_job.job)
+                    else:
+                        # still running
+                        still_running.append(active_job)
+
+                        def handle_suspended(by):
+                            if active_job.job.jobid not in suspended_msg:
+                                logger.warning(
+                                    "Job {} (DRMAA id: {}) was suspended by {}.".format(
+                                        active_job.job.jobid, active_job.jobid, by
+                                    )
+                                )
+                                suspended_msg.add(active_job.job.jobid)
+
+                        if retval == drmaa.JobState.USER_SUSPENDED:
+                            handle_suspended("user")
+                        elif retval == drmaa.JobState.SYSTEM_SUSPENDED:
+                            handle_suspended("system")
+                        else:
+                            try:
+                                suspended_msg.remove(active_job.job.jobid)
+                            except KeyError:
+                                # there was nothing to remove
+                                pass
+
             with self.lock:
                 self.active_jobs.extend(still_running)
             sleep()
@@ -1323,7 +1470,7 @@ class DRMAAExecutor(ClusterExecutor):
 
 @contextlib.contextmanager
 def change_working_directory(directory=None):
-    """ Change working directory in execution context if provided. """
+    """Change working directory in execution context if provided."""
     if directory:
         try:
             saved_directory = os.getcwd()
@@ -1357,14 +1504,16 @@ class KubernetesExecutor(ClusterExecutor):
         local_input=None,
         restart_times=None,
         keepincomplete=False,
+        keepmetadata=True,
     ):
+        self.workflow = workflow
 
         exec_job = (
             "cp -rf /source/. . && "
             "snakemake {target} --snakefile {snakefile} "
-            "--force -j{cores} --keep-target-files  --keep-remote "
-            "--latency-wait {latency_wait} "
-            " --attempt {attempt} {use_threads} "
+            "--force --cores {cores} --keep-target-files  --keep-remote "
+            "--latency-wait {latency_wait} --scheduler {workflow.scheduler_type} "
+            " --attempt {attempt} {use_threads} --max-inventory-time 0 "
             "--wrapper-prefix {workflow.wrapper_prefix} "
             "{overwrite_config} {printshellcmds} {rules} --nocolor "
             "--notemp --no-hooks --nolock "
@@ -1387,7 +1536,7 @@ class KubernetesExecutor(ClusterExecutor):
             max_status_checks_per_second=10,
         )
         # use relative path to Snakefile
-        self.snakefile = os.path.relpath(workflow.snakefile)
+        self.snakefile = os.path.relpath(workflow.main_snakefile)
 
         try:
             from kubernetes import config
@@ -1429,7 +1578,7 @@ class KubernetesExecutor(ClusterExecutor):
 
             # The kubernetes API can't create secret files larger than 1MB.
             source_file_size = os.path.getsize(f)
-            max_file_size = 1000000
+            max_file_size = 1048576
             if source_file_size > max_file_size:
                 logger.warning(
                     "Skipping the source file {f}. Its size {source_file_size} exceeds "
@@ -1442,8 +1591,28 @@ class KubernetesExecutor(ClusterExecutor):
 
             with open(f, "br") as content:
                 key = "f{}".format(i)
+
+                # Some files are smaller than 1MB, but grows larger after being base64 encoded
+                # We should exclude them as well, otherwise Kubernetes APIs will complain
+                encoded_contents = base64.b64encode(content.read()).decode()
+                encoded_size = len(encoded_contents)
+                if encoded_size > 1048576:
+                    logger.warning(
+                        "Skipping the source file {f} for secret key {key}. "
+                        "Its base64 encoded size {encoded_size} exceeds "
+                        "the maximum file size (1MB) that can be passed "
+                        "from host to kubernetes.".format(
+                            f=f,
+                            source_file_size=source_file_size,
+                            key=key,
+                            encoded_size=encoded_size,
+                        )
+                    )
+                    continue
+
                 self.secret_files[key] = f
-                secret.data[key] = base64.b64encode(content.read()).decode()
+                secret.data[key] = encoded_contents
+
         for e in self.envvars:
             try:
                 key = e.lower()
@@ -1451,14 +1620,58 @@ class KubernetesExecutor(ClusterExecutor):
                 self.secret_envvars[key] = e
             except KeyError:
                 continue
+
+        # Test if the total size of the configMap exceeds 1MB
+        config_map_size = sum(
+            [len(base64.b64decode(v)) for k, v in secret.data.items()]
+        )
+        if config_map_size > 1048576:
+            logger.warning(
+                "The total size of the included files and other Kubernetes secrets "
+                "is {}, exceeding the 1MB limit.\n".format(config_map_size)
+            )
+            logger.warning(
+                "The following are the largest files. Consider removing some of them "
+                "(you need remove at least {} bytes):".format(config_map_size - 1048576)
+            )
+
+            entry_sizes = {
+                self.secret_files[k]: len(base64.b64decode(v))
+                for k, v in secret.data.items()
+                if k in self.secret_files
+            }
+            for k, v in sorted(entry_sizes.items(), key=lambda item: item[1])[:-6:-1]:
+                logger.warning("  * File: {k}, original size: {v}".format(k=k, v=v))
+
+            raise WorkflowError("ConfigMap too large")
+
         self.kubeapi.create_namespaced_secret(self.namespace, secret)
 
     def unregister_secret(self):
         import kubernetes.client
 
-        self.kubeapi.delete_namespaced_secret(
+        safe_delete_secret = lambda: self.kubeapi.delete_namespaced_secret(
             self.run_namespace, self.namespace, body=kubernetes.client.V1DeleteOptions()
         )
+        self._kubernetes_retry(safe_delete_secret)
+
+    # In rare cases, deleting a pod may rais 404 NotFound error.
+    def safe_delete_pod(self, jobid, ignore_not_found=True):
+        import kubernetes.client
+
+        body = kubernetes.client.V1DeleteOptions()
+        try:
+            self.kubeapi.delete_namespaced_pod(jobid, self.namespace, body=body)
+        except kubernetes.client.rest.ApiException as e:
+            if e.status == 404 and ignore_not_found:
+                # Can't find the pod. Maybe it's already been
+                # destroyed. Proceed with a warning message.
+                logger.warning(
+                    "[WARNING] 404 not found when trying to delete the pod: {jobid}\n"
+                    "[WARNING] Ignore this error\n".format(jobid=jobid)
+                )
+            else:
+                raise e
 
     def shutdown(self):
         self.unregister_secret()
@@ -1470,7 +1683,9 @@ class KubernetesExecutor(ClusterExecutor):
         body = kubernetes.client.V1DeleteOptions()
         with self.lock:
             for j in self.active_jobs:
-                self.kubeapi.delete_namespaced_pod(j.jobid, self.namespace, body=body)
+                func = lambda: self.safe_delete_pod(j.jobid, ignore_not_found=True)
+                self._kubernetes_retry(func)
+
         self.shutdown()
 
     def run(self, job, callback=None, submit_callback=None, error_callback=None):
@@ -1491,6 +1706,7 @@ class KubernetesExecutor(ClusterExecutor):
 
         body = kubernetes.client.V1Pod()
         body.metadata = kubernetes.client.V1ObjectMeta(labels={"app": "snakemake"})
+
         body.metadata.name = jobid
 
         # container
@@ -1588,6 +1804,43 @@ class KubernetesExecutor(ClusterExecutor):
             KubernetesJob(job, jobid, callback, error_callback, pod, None)
         )
 
+    # Sometimes, certain k8s requests throw kubernetes.client.rest.ApiException
+    # Solving this issue requires reauthentication, as _kubernetes_retry shows
+    # However, reauthentication itself, under rare conditions, may also throw
+    # errors such as:
+    #   kubernetes.client.exceptions.ApiException: (409), Reason: Conflict
+    #
+    # This error doesn't mean anything wrong with the k8s cluster, and users can safely
+    # ignore it.
+    def _reauthenticate_and_retry(self, func=None):
+        import kubernetes
+
+        # Unauthorized.
+        # Reload config in order to ensure token is
+        # refreshed. Then try again.
+        logger.info("Trying to reauthenticate")
+        kubernetes.config.load_kube_config()
+        subprocess.run(["kubectl", "get", "nodes"])
+
+        self.kubeapi = kubernetes.client.CoreV1Api()
+        self.batchapi = kubernetes.client.BatchV1Api()
+
+        try:
+            self.register_secret()
+        except kubernetes.client.rest.ApiException as e:
+            if e.status == 409 and e.reason == "Conflict":
+                logger.warning("409 conflict ApiException when registering secrets")
+                logger.warning(e)
+            else:
+                raise WorkflowError(
+                    e,
+                    "This is likely a bug in "
+                    "https://github.com/kubernetes-client/python.",
+                )
+
+        if func:
+            return func()
+
     def _kubernetes_retry(self, func):
         import kubernetes
         import urllib3
@@ -1600,22 +1853,7 @@ class KubernetesExecutor(ClusterExecutor):
                     # Unauthorized.
                     # Reload config in order to ensure token is
                     # refreshed. Then try again.
-                    logger.info("trying to reauthenticate")
-                    kubernetes.config.load_kube_config()
-                    subprocess.run(["kubectl", "get", "nodes"])
-
-                    self.kubeapi = kubernetes.client.CoreV1Api()
-                    self.batchapi = kubernetes.client.BatchV1Api()
-                    self.register_secret()
-                    try:
-                        return func()
-                    except kubernetes.client.rest.ApiException as e:
-                        # Both attempts failed, raise error.
-                        raise WorkflowError(
-                            e,
-                            "This is likely a bug in "
-                            "https://github.com/kubernetes-client/python.",
-                        )
+                    return self._reauthenticate_and_retry(func)
             # Handling timeout that may occur in case of GKE master upgrade
             except urllib3.exceptions.MaxRetryError as e:
                 logger.info(
@@ -1686,10 +1924,11 @@ class KubernetesExecutor(ClusterExecutor):
                     elif res.status.phase == "Succeeded":
                         # finished
                         j.callback(j.job)
-                        body = kubernetes.client.V1DeleteOptions()
-                        self.kubeapi.delete_namespaced_pod(
-                            j.jobid, self.namespace, body=body
+
+                        func = lambda: self.safe_delete_pod(
+                            j.jobid, ignore_not_found=True
                         )
+                        self._kubernetes_retry(func)
                     else:
                         # still active
                         still_running.append(j)
@@ -1721,8 +1960,9 @@ class TibannaExecutor(ClusterExecutor):
         restart_times=None,
         max_status_checks_per_second=1,
         keepincomplete=False,
+        keepmetadata=True,
     ):
-
+        self.workflow = workflow
         self.workflow_sources = []
         for wfs in workflow.get_sources():
             if os.path.isdir(wfs):
@@ -1737,7 +1977,7 @@ class TibannaExecutor(ClusterExecutor):
         for f in self.workflow_sources:
             log += f
         logger.debug(log)
-        self.snakefile = workflow.snakefile
+        self.snakefile = workflow.main_snakefile
         self.envvars = {e: os.environ[e] for e in workflow.envvars}
         if self.envvars:
             logger.debug("envvars = %s" % str(self.envvars))
@@ -1756,9 +1996,9 @@ class TibannaExecutor(ClusterExecutor):
         self.quiet = quiet
         exec_job = (
             "snakemake {target} --snakefile {snakefile} "
-            "--force -j{cores} --keep-target-files  --keep-remote "
-            "--latency-wait 0 "
-            "--attempt 1 {use_threads} "
+            "--force --cores {cores} --keep-target-files  --keep-remote "
+            "--latency-wait 0 --scheduler {workflow.scheduler_type} "
+            "--attempt 1 {use_threads} --max-inventory-time 0 "
             "{overwrite_config} {rules} --nocolor "
             "--notemp --no-hooks --nolock "
         )
@@ -1986,6 +2226,7 @@ class TibannaExecutor(ClusterExecutor):
             sfn=self.tibanna_sfn,
             verbose=not self.quiet,
             jobid=jobid,
+            open_browser=False,
             sleep=0,
         )
         exec_arn = exec_info.get("_tibanna", {}).get("exec_arn", "")
@@ -2055,6 +2296,9 @@ def run_wrapper(
     shadow_dir,
     jobid,
     edit_notebook,
+    conda_base_path,
+    basedir,
+    runtime_sourcecache_path,
 ):
     """
     Wrapper around the run method that handles exceptions and benchmarking.
@@ -2133,6 +2377,9 @@ def run_wrapper(
                             cleanup_scripts,
                             passed_shadow_dir,
                             edit_notebook,
+                            conda_base_path,
+                            basedir,
+                            runtime_sourcecache_path,
                         )
                     else:
                         # The benchmarking is started here as we have a run section
@@ -2161,6 +2408,9 @@ def run_wrapper(
                                 cleanup_scripts,
                                 passed_shadow_dir,
                                 edit_notebook,
+                                conda_base_path,
+                                basedir,
+                                runtime_sourcecache_path,
                             )
                     # Store benchmark record for this iteration
                     bench_records.append(bench_record)
@@ -2187,20 +2437,28 @@ def run_wrapper(
                     cleanup_scripts,
                     passed_shadow_dir,
                     edit_notebook,
+                    conda_base_path,
+                    basedir,
+                    runtime_sourcecache_path,
                 )
     except (KeyboardInterrupt, SystemExit) as e:
         # Re-raise the keyboard interrupt in order to record an error in the
         # scheduler but ignore it
         raise e
     except (Exception, BaseException) as ex:
-        log_verbose_traceback(ex)
         # this ensures that exception can be re-raised in the parent thread
-        lineno, file = get_exception_origin(ex, linemaps)
-        raise RuleException(
-            format_error(
-                ex, lineno, linemaps=linemaps, snakefile=file, show_traceback=True
+        origin = get_exception_origin(ex, linemaps)
+        if origin is not None:
+            log_verbose_traceback(ex)
+            lineno, file = origin
+            raise RuleException(
+                format_error(
+                    ex, lineno, linemaps=linemaps, snakefile=file, show_traceback=True
+                )
             )
-        )
+        else:
+            # some internal bug, just reraise
+            raise ex
 
     if benchmark is not None:
         try:
