@@ -1,6 +1,6 @@
 __author__ = "Johannes Köster"
-__copyright__ = "Copyright 2015-2019, Johannes Köster"
-__email__ = "koester@jimmy.harvard.edu"
+__copyright__ = "Copyright 2021, Johannes Köster"
+__email__ = "johannes.koester@uni-due.de"
 __license__ = "MIT"
 
 import tokenize
@@ -11,6 +11,7 @@ import urllib.request
 from io import TextIOWrapper
 
 from snakemake.exceptions import WorkflowError
+from snakemake import common
 
 dd = textwrap.dedent
 
@@ -122,7 +123,7 @@ class TokenAutomaton:
             base_indent=self.base_indent + self.indent,
             dedent=self.dedent,
             root=False,
-            **kwargs
+            **kwargs,
         )
 
 
@@ -144,7 +145,10 @@ class KeywordState(TokenAutomaton):
 
     def decorate_end(self, token):
         for t in self.end():
-            yield t, token
+            if isinstance(t, tuple):
+                yield t
+            else:
+                yield t, token
 
     def colon(self, token):
         if is_colon(token):
@@ -162,8 +166,7 @@ class KeywordState(TokenAutomaton):
             # ignore lines containing only comments
             self.line -= 1
         if self.is_block_end(token):
-            for t, token_ in self.decorate_end(token):
-                yield t, token_
+            yield from self.decorate_end(token)
             yield "\n", token
             raise StopAutomaton(token)
 
@@ -174,8 +177,7 @@ class KeywordState(TokenAutomaton):
             if is_comment(token):
                 yield token.string, token
             else:
-                for t in self.block_content(token):
-                    yield t
+                yield from self.block_content(token)
 
     def yield_indent(self, token):
         return token.string, token
@@ -257,6 +259,10 @@ class Report(GlobalKeywordState):
     pass
 
 
+class Scattergather(GlobalKeywordState):
+    pass
+
+
 class Ruleorder(GlobalKeywordState):
     def block_content(self, token):
         if is_greater(token):
@@ -289,6 +295,12 @@ class GlobalContainer(GlobalKeywordState):
         return "global_container"
 
 
+class GlobalContainerized(GlobalKeywordState):
+    @property
+    def keyword(self):
+        return "global_containerized"
+
+
 # subworkflows
 
 
@@ -392,6 +404,10 @@ class Localrules(GlobalKeywordState):
 # Rule keyword states
 
 
+class Name(RuleKeywordState):
+    pass
+
+
 class Input(RuleKeywordState):
     pass
 
@@ -450,6 +466,10 @@ class Container(RuleKeywordState):
     pass
 
 
+class Containerized(RuleKeywordState):
+    pass
+
+
 class EnvModules(RuleKeywordState):
     pass
 
@@ -464,6 +484,10 @@ class Cache(RuleKeywordState):
         return "cache_rule"
 
 
+class Handover(RuleKeywordState):
+    pass
+
+
 class WildcardConstraints(RuleKeywordState):
     @property
     def keyword(self):
@@ -483,10 +507,12 @@ class Run(RuleKeywordState):
             "def __rule_{rulename}(input, output, params, wildcards, threads, "
             "resources, log, version, rule, conda_env, container_img, "
             "singularity_args, use_singularity, env_modules, bench_record, jobid, "
-            "is_shell, bench_iteration, cleanup_scripts, shadow_dir, edit_notebook):".format(
+            "is_shell, bench_iteration, cleanup_scripts, shadow_dir, edit_notebook, "
+            "conda_base_path, basedir, runtime_sourcecache_path, {rule_func_marker}=True):".format(
                 rulename=self.rulename
                 if self.rulename is not None
-                else self.snakefile.rulecount
+                else self.snakefile.rulecount,
+                rule_func_marker=common.RULEFUNC_CONTEXT_MARKER,
             )
         )
 
@@ -581,13 +607,10 @@ class Script(AbstractCmd):
     end_func = "script"
 
     def args(self):
-        # basedir
-        yield ", {!r}".format(os.path.abspath(os.path.dirname(self.snakefile.path)))
-        # other args
         yield (
-            ", input, output, params, wildcards, threads, resources, log, "
-            "config, rule, conda_env, container_img, singularity_args, env_modules, "
-            "bench_record, jobid, bench_iteration, cleanup_scripts, shadow_dir"
+            ", basedir, input, output, params, wildcards, threads, resources, log, "
+            "config, rule, conda_env, conda_base_path, container_img, singularity_args, env_modules, "
+            "bench_record, jobid, bench_iteration, cleanup_scripts, shadow_dir, runtime_sourcecache_path"
         )
 
 
@@ -596,14 +619,11 @@ class Notebook(Script):
     end_func = "notebook"
 
     def args(self):
-        # basedir
-        yield ", {!r}".format(os.path.abspath(os.path.dirname(self.snakefile.path)))
-        # other args
         yield (
-            ", input, output, params, wildcards, threads, resources, log, "
-            "config, rule, conda_env, container_img, singularity_args, env_modules, "
+            ", basedir, input, output, params, wildcards, threads, resources, log, "
+            "config, rule, conda_env, conda_base_path, container_img, singularity_args, env_modules, "
             "bench_record, jobid, bench_iteration, cleanup_scripts, shadow_dir, "
-            "edit_notebook"
+            "edit_notebook, runtime_sourcecache_path"
         )
 
 
@@ -614,9 +634,9 @@ class Wrapper(Script):
     def args(self):
         yield (
             ", input, output, params, wildcards, threads, resources, log, "
-            "config, rule, conda_env, container_img, singularity_args, env_modules, "
+            "config, rule, conda_env, conda_base_path, container_img, singularity_args, env_modules, "
             "bench_record, workflow.wrapper_prefix, jobid, bench_iteration, "
-            "cleanup_scripts, shadow_dir"
+            "cleanup_scripts, shadow_dir, runtime_sourcecache_path"
         )
 
 
@@ -625,41 +645,46 @@ class CWL(Script):
     end_func = "cwl"
 
     def args(self):
-        # basedir
-        yield ", {!r}".format(os.path.abspath(os.path.dirname(self.snakefile.path)))
-        # other args
         yield (
-            ", input, output, params, wildcards, threads, resources, log, "
-            "config, rule, use_singularity, bench_record, jobid"
+            ", basedir, input, output, params, wildcards, threads, resources, log, "
+            "config, rule, use_singularity, bench_record, jobid, runtime_sourcecache_path"
         )
 
 
+rule_property_subautomata = dict(
+    name=Name,
+    input=Input,
+    output=Output,
+    params=Params,
+    threads=Threads,
+    resources=Resources,
+    priority=Priority,
+    version=Version,
+    log=Log,
+    message=Message,
+    benchmark=Benchmark,
+    conda=Conda,
+    singularity=Singularity,
+    container=Container,
+    containerized=Containerized,
+    envmodules=EnvModules,
+    wildcard_constraints=WildcardConstraints,
+    shadow=Shadow,
+    group=Group,
+    cache=Cache,
+    handover=Handover,
+)
+
+
 class Rule(GlobalKeywordState):
     subautomata = dict(
-        input=Input,
-        output=Output,
-        params=Params,
-        threads=Threads,
-        resources=Resources,
-        priority=Priority,
-        version=Version,
-        log=Log,
-        message=Message,
-        benchmark=Benchmark,
-        conda=Conda,
-        singularity=Singularity,
-        container=Container,
-        envmodules=EnvModules,
-        wildcard_constraints=WildcardConstraints,
-        shadow=Shadow,
-        group=Group,
         run=Run,
         shell=Shell,
         script=Script,
         notebook=Notebook,
         wrapper=Wrapper,
         cwl=CWL,
-        cache=Cache,
+        **rule_property_subautomata,
     )
 
     def __init__(self, snakefile, base_indent=0, dedent=0, root=True):
@@ -783,6 +808,308 @@ class OnStart(DecoratorKeywordState):
     args = ["log"]
 
 
+# modules
+
+
+class ModuleKeywordState(SectionKeywordState):
+    prefix = "Module"
+
+
+class ModuleSnakefile(ModuleKeywordState):
+    pass
+
+
+class ModuleMetaWrapper(ModuleKeywordState):
+    @property
+    def keyword(self):
+        return "meta_wrapper"
+
+
+class ModuleConfig(ModuleKeywordState):
+    pass
+
+
+class ModuleSkipValidation(ModuleKeywordState):
+    @property
+    def keyword(self):
+        return "skip_validation"
+
+
+class ModuleReplacePrefix(ModuleKeywordState):
+    @property
+    def keyword(self):
+        return "replace_prefix"
+
+
+class Module(GlobalKeywordState):
+    subautomata = dict(
+        snakefile=ModuleSnakefile,
+        meta_wrapper=ModuleMetaWrapper,
+        config=ModuleConfig,
+        skip_validation=ModuleSkipValidation,
+        replace_prefix=ModuleReplacePrefix,
+    )
+
+    def __init__(self, snakefile, base_indent=0, dedent=0, root=True):
+        super().__init__(snakefile, base_indent=base_indent, dedent=dedent, root=root)
+        self.state = self.name
+        self.has_snakefile = False
+        self.has_meta_wrapper = False
+        self.has_name = False
+        self.primary_token = None
+
+    def end(self):
+        if not (self.has_snakefile or self.has_meta_wrapper):
+            self.error(
+                "A module needs either a path to a Snakefile or a meta wrapper URL.",
+                self.primary_token,
+            )
+        yield ")"
+
+    def name(self, token):
+        if is_name(token):
+            yield "workflow.module({name!r}".format(name=token.string), token
+            self.has_name = True
+        elif is_colon(token) and self.has_name:
+            self.primary_token = token
+            self.state = self.block
+        else:
+            self.error("Expected name after module keyword.", token)
+
+    def block_content(self, token):
+        if is_name(token):
+            try:
+                if token.string == "snakefile":
+                    self.has_snakefile = True
+                if token.string == "meta_wrapper":
+                    self.has_meta_wrapper = True
+                for t in self.subautomaton(token.string).consume():
+                    yield t
+            except KeyError:
+                self.error(
+                    "Unexpected keyword {} in "
+                    "module definition".format(token.string),
+                    token,
+                )
+            except StopAutomaton as e:
+                self.indentation(e.token)
+                for t in self.block(e.token):
+                    yield t
+        elif is_comment(token):
+            yield "\n", token
+            yield token.string, token
+        elif is_string(token):
+            # ignore docstring
+            pass
+        else:
+            self.error(
+                "Expecting module keyword, comment or docstrings "
+                "inside a module definition.",
+                token,
+            )
+
+
+class UseRule(GlobalKeywordState):
+    subautomata = rule_property_subautomata
+
+    def __init__(self, snakefile, base_indent=0, dedent=0, root=True):
+        super().__init__(snakefile, base_indent=base_indent, dedent=dedent, root=root)
+        self.state = self.state_keyword_rule
+        self.rules = []
+        self.has_with = False
+        self.name_modifier = []
+        self.from_module = None
+        self._with_block = []
+        self.lineno = self.snakefile.lines + 1
+
+    def end(self):
+        name_modifier = "".join(self.name_modifier) if self.name_modifier else None
+        yield "@workflow.userule(rules={!r}, from_module={!r}, name_modifier={!r}, lineno={})".format(
+            self.rules, self.from_module, name_modifier, self.lineno
+        )
+        yield "\n"
+
+        # yield with block
+        yield from self._with_block
+
+        yield "@workflow.run"
+        yield "\n"
+
+        rulename = self.rules[0]
+        if rulename == "*":
+            rulename = "__allrules__"
+        yield "def __userule_{}_{}():".format(self.from_module, rulename)
+        # the end is detected.
+        # So we can savely reset the indent to zero here
+        self.indent = 0
+        yield "\n"
+        yield INDENT * (self.effective_indent + 1)
+        yield "pass"
+
+    def state_keyword_rule(self, token):
+        if is_name(token) and token.string == "rule":
+            self.state = self.state_rules_rule
+            yield from ()
+        else:
+            self.error("Expecting keyword 'rule' after keyword 'use'", token)
+
+    def state_rules_rule(self, token):
+        if is_name(token):
+            if token.string == "from" or token.string == "as" and not self.rules:
+                self.error("Expecting rule names after 'use rule' statement.", token)
+
+            self.rules.append(token.string)
+            self.state = self.state_rules_comma_or_end
+            yield from ()
+        elif is_op(token):
+            if token.string == "*":
+                self.rules.append(token.string)
+                self.state = self.state_rules_end
+                yield from ()
+            else:
+                self.error(
+                    "Expecting rule name or '*' after 'use rule' statement.", token
+                )
+        else:
+            self.error(
+                "Expecting rule listing (comma separated) after 'use rule' statement.",
+                token,
+            )
+        # TODO newline and parentheses handling
+
+    def state_rules_end(self, token):
+        if is_name(token) and token.string == "from":
+            self.state = self.state_from
+            yield from ()
+        else:
+            self.error(
+                "Expecting list of rules in 'use rule' statement to end with keyword 'from'.",
+                token,
+            )
+
+    def state_rules_comma_or_end(self, token):
+        if is_name(token):
+            if token.string == "from" or token.string == "as":
+                if not self.rules:
+                    self.error(
+                        "Expecting rule names after 'use rule' statement.", token
+                    )
+                if token.string == "from":
+                    self.state = self.state_from
+                else:
+                    self.state = self.state_as
+                yield from ()
+            else:
+                self.error(
+                    "Expecting list of rules in 'use rule' statement to end with keyword 'from'.",
+                    token,
+                )
+        elif is_comma(token):
+            self.state = self.state_rules_rule
+            yield from ()
+        else:
+            self.error(
+                "Unexpected token in list of rules within 'use rule' statement.", token
+            )
+
+    def state_from(self, token):
+        if is_name(token):
+            self.state = self.state_modifier
+            self.from_module = token.string
+            yield from ()
+        else:
+            self.error(
+                "Expecting module name after 'from' keyword in 'use rule' statement.",
+                token,
+            )
+
+    def state_modifier(self, token):
+        if is_name(token):
+            if token.string == "as" and not self.name_modifier:
+                self.state = self.state_as
+                yield from ()
+            elif token.string == "with":
+                yield from self.handle_with(token)
+            else:
+                self.error(
+                    "Expecting at most one 'as' or 'with' statement, or the end of the line.",
+                    token,
+                )
+        elif is_newline(token) or is_comment(token) or is_eof(token):
+            # end of the statement, close block manually
+            yield from self.block(token)
+        else:
+            self.error(
+                "Expecting either 'as', 'with' or end of line in 'use rule' statement.",
+                token,
+            )
+
+    def handle_with(self, token):
+        if "*" in self.rules:
+            self.error(
+                "Keyword 'with' in 'use rule' statement is not allowed in combination with rule pattern '*'.",
+                token,
+            )
+        self.has_with = True
+        self.state = self.state_with
+        yield from ()
+
+    def state_as(self, token):
+        if is_name(token):
+            if token.string != "with":
+                self.name_modifier.append(token.string)
+                yield from ()
+            else:
+                yield from self.handle_with(token)
+        elif is_op(token) and token.string == "*":
+            self.name_modifier.append(token.string)
+            yield from ()
+        elif is_newline(token) or is_comment(token) or is_eof(token):
+            # end of the statement, close block manually
+            yield from self.block(token)
+        else:
+            self.error(
+                "Expecting rulename modifying pattern (e.g. modulename_*) after 'as' keyword.",
+                token,
+            )
+
+    def state_with(self, token):
+        if is_colon(token):
+            self.state = self.block
+            yield from ()
+        else:
+            self.error(
+                "Expecting colon after 'with' keyword in 'use rule' statement.", token
+            )
+
+    def block_content(self, token):
+        if is_comment(token):
+            yield "\n", token
+            yield token.string, token
+        elif is_name(token):
+            try:
+                self._with_block.extend(self.subautomaton(token.string).consume())
+                yield from ()
+            except KeyError:
+                self.error(
+                    "Unexpected keyword {} in rule definition".format(token.string),
+                    token,
+                )
+            except StopAutomaton as e:
+                self.indentation(e.token)
+                self.block(e.token)
+        else:
+            self.error(
+                "Expecting a keyword or comment "
+                "inside a 'use rule ... with:' statement.",
+                token,
+            )
+
+    @property
+    def dedent(self):
+        return self.indent
+
+
 class Python(TokenAutomaton):
 
     subautomata = dict(
@@ -804,6 +1131,10 @@ class Python(TokenAutomaton):
         wildcard_constraints=GlobalWildcardConstraints,
         singularity=GlobalSingularity,
         container=GlobalContainer,
+        containerized=GlobalContainerized,
+        scattergather=Scattergather,
+        module=Module,
+        use=UseRule,
     )
 
     def __init__(self, snakefile, base_indent=0, dedent=0, root=True):
@@ -827,18 +1158,9 @@ class Python(TokenAutomaton):
 
 
 class Snakefile:
-    def __init__(self, path, rulecount=0):
-        self.path = path
-        try:
-            self.file = open(self.path, encoding="utf-8")
-        except (FileNotFoundError, OSError) as e:
-            try:
-                self.file = TextIOWrapper(
-                    urllib.request.urlopen(self.path), encoding="utf-8"
-                )
-            except (HTTPError, URLError, ContentTooShortError, ValueError):
-                raise WorkflowError("Failed to open {}.".format(path))
-
+    def __init__(self, path, workflow, rulecount=0):
+        self.path = path.get_path_or_uri()
+        self.file = workflow.sourcecache.open(path)
         self.tokens = tokenize.generate_tokens(self.file.readline)
         self.rulecount = rulecount
         self.lines = 0
@@ -865,9 +1187,9 @@ def format_tokens(tokens):
         t_ = t
 
 
-def parse(path, overwrite_shellcmd=None, rulecount=0):
+def parse(path, workflow, overwrite_shellcmd=None, rulecount=0):
     Shell.overwrite_cmd = overwrite_shellcmd
-    with Snakefile(path, rulecount=rulecount) as snakefile:
+    with Snakefile(path, workflow, rulecount=rulecount) as snakefile:
         automaton = Python(snakefile)
         linemap = dict()
         compilation = list()
